from marketsim import Market, Order, OrderStat, Side
import re
import unittest

class TestOrderBook(unittest.TestCase):
    def strip_spaces(self, str):
        str = re.sub(r'^\s+|\s+$', '', str)
        str = re.compile(r'^\s+', re.MULTILINE).sub('', str)
        return str

    def test_order_book_entry(self):
        market = Market()
        market.place(Order(Side.BUY, 'abc', 10, 110))
        market.place(Order(Side.SELL, 'abc', 10, 130))
        bid_order_book = market['abc'][Side.BUY].get_order_book()
        ask_order_book = market['abc'][Side.SELL].get_order_book()
        self.assertEqual(bid_order_book, [OrderStat(110, 10, 1)])
        self.assertEqual(ask_order_book, [OrderStat(130, 10, 1)])
        self.assertEqual(repr(bid_order_book[0]), 'OrderStat(price=110, volume=10, count=1)')
        self.assertEqual(repr(ask_order_book[0]), 'OrderStat(price=130, volume=10, count=1)')

    def test_side_order_book(self):
        market = Market()

        for price in [110, 120, 130, 140]:
            market.place(Order(Side.SELL, 'abc', 10, price))
            market.place(Order(Side.SELL, 'abc', 10, price))

        for price in [100, 110, 120, 130]:
            market.place(Order(Side.BUY, 'abc', 10, price))
            market.place(Order(Side.BUY, 'abc', 10, price))

    def test_format_order_book(self):
        market = Market()

        for price in [110, 120, 130, 140]:
            market.place(Order(Side.SELL, 'abc', 10, price))
            market.place(Order(Side.SELL, 'abc', 10, price))

        for price in [100, 110, 120, 130]:
            market.place(Order(Side.BUY, 'abc', 10, price))
            market.place(Order(Side.BUY, 'abc', 10, price))

        self.assertEqual(market['abc'].format_order_book(), self.strip_spaces(
            """
            | BID    | PRICE | ASK    |
            |========|=======|========|
            |        | 140   | 20 (2) |
            | 20 (2) | 130   | 20 (2) |
            | 20 (2) | 120   | 20 (2) |
            | 20 (2) | 110   | 20 (2) |
            | 20 (2) | 100   |        |
            """
        ))

        market.execute()

        self.assertEqual(market['abc'].format_order_book(), self.strip_spaces(
            """
            | BID    | PRICE | ASK    |
            |========|=======|========|
            |        | 140   | 20 (2) |
            |        | 130   | 20 (2) |
            | 20 (2) | 110   |        |
            | 20 (2) | 100   |        |
            """
        ))

    def test_order_book_after_cancel(self):
        market = Market()

        for price in [110, 120, 130, 140]:
            market.place(Order(Side.SELL, 'abc', 10, price, id='sell-{}-1'.format(price)))
            market.place(Order(Side.SELL, 'abc', 10, price, id='sell-{}-2'.format(price)))

        for price in [100, 110, 120, 130]:
            market.place(Order(Side.BUY, 'abc', 10, price, id='buy-{}-1'.format(price)))
            market.place(Order(Side.BUY, 'abc', 10, price, id='buy-{}-2'.format(price)))

        market.cancel(Order(id='buy-120-1', symbol='abc'))
        market.cancel(Order(id='sell-130-1', symbol='abc'))

        self.assertEqual(market['abc'].format_order_book(), self.strip_spaces(
            """
            | BID    | PRICE | ASK    |
            |========|=======|========|
            |        | 140   | 20 (2) |
            | 20 (2) | 130   | 10 (1) |
            | 10 (1) | 120   | 20 (2) |
            | 20 (2) | 110   | 20 (2) |
            | 20 (2) | 100   |        |
            """
        ))

        market.cancel(Order(id='buy-120-2', symbol='abc'))
        market.cancel(Order(id='sell-130-2', symbol='abc'))

        self.assertEqual(market['abc'].format_order_book(), self.strip_spaces(
            """
            | BID    | PRICE | ASK    |
            |========|=======|========|
            |        | 140   | 20 (2) |
            | 20 (2) | 130   |        |
            |        | 120   | 20 (2) |
            | 20 (2) | 110   | 20 (2) |
            | 20 (2) | 100   |        |
            """
        ))



from marketsim import Market, Order, Side, Fill, Execution, Allocation, OrderEntry
import unittest

class TestOrder(unittest.TestCase):
    def test_side_enum(self):
        self.assertEqual(str(Side.BUY), 'Side.BUY')
        self.assertEqual(str(Side.SELL), 'Side.SELL')
        self.assertEqual(repr(Side.BUY), '<Side.BUY: 1>')
        self.assertEqual(repr(Side.SELL), '<Side.SELL: 2>')

        self.assertEqual(Side.normalize(None), None)
        self.assertEqual(Side.normalize(1), Side.BUY)
        self.assertEqual(Side.normalize(2), Side.SELL)
        self.assertEqual(Side.normalize('BUY'), Side.BUY)
        self.assertEqual(Side.normalize('SELL'), Side.SELL)
        self.assertEqual(Side.normalize('buy'), Side.BUY)
        self.assertEqual(Side.normalize('sell'), Side.SELL)

        with self.assertRaises(KeyError):
            Side.normalize('UNKNOWN')

        with self.assertRaises(ValueError):
            Side.normalize(-1)

        with self.assertRaises(ValueError):
            Side.normalize(1.0)

    def test_objects(self):
        bid_order = Order(Side.BUY, 'abc', 10)
        ask_order = Order(Side.SELL, 'abc', 10)
        self.assertEqual(repr(bid_order), 'Order(side=Side.BUY, symbol=abc, quantity=10, price=None)')
        self.assertEqual(repr(ask_order), 'Order(side=Side.SELL, symbol=abc, quantity=10, price=None)')

        bid_entry = OrderEntry(bid_order)
        ask_entry = OrderEntry(ask_order)
        execution = Execution(bid_entry, ask_entry, 5, 100)
        bid_fill = Fill(bid_entry, 5, 100)
        ask_fill = Fill(ask_entry, 5, 100)
        self.assertEqual(repr(execution), 'Execution(bid_fill={}, ask_fill={}, quantity=5, price=100)'.format(bid_fill, ask_fill))

        allocation = Allocation(bid_order, 5)
        self.assertEqual(repr(allocation), 'Allocation(entry={}, quantity=5)'.format(bid_order))

    def test_equal_buy_sell(self):
        bid_entry = OrderEntry(Order(Side.BUY, 'abc', 10, 130))
        ask_entry = OrderEntry(Order(Side.SELL, 'abc', 10, 110))

        execution = bid_entry.execute(ask_entry)

        self.assertEqual(execution.price, None)
        self.assertEqual(execution.quantity, 10)

        self.assertEqual(bid_entry.quantity, 10)
        self.assertEqual(ask_entry.quantity, 10)
        self.assertEqual(bid_entry.remaining, 0)
        self.assertEqual(ask_entry.remaining, 0)
        self.assertEqual(bid_entry.filled_quantity, 10)
        self.assertEqual(ask_entry.filled_quantity, 10)

        self.assertIs(execution.bid_fill.order, bid_entry.order)
        self.assertIs(execution.ask_fill.order, ask_entry.order)

    def test_more_buy_than_sell(self):
        bid_entry = OrderEntry(Order(Side.BUY, 'abc', 10, 130))
        ask_entry = OrderEntry(Order(Side.SELL, 'abc', 15, 110))

        execution = bid_entry.execute(ask_entry)

        self.assertEqual(execution.price, None)
        self.assertEqual(execution.quantity, 10)

        self.assertEqual(bid_entry.quantity, 10)
        self.assertEqual(ask_entry.quantity, 15)
        self.assertEqual(bid_entry.remaining, 0)
        self.assertEqual(ask_entry.remaining, 5)
        self.assertEqual(bid_entry.filled_quantity, 10)
        self.assertEqual(ask_entry.filled_quantity, 10)

        self.assertIs(execution.bid_fill.order, bid_entry.order)
        self.assertIs(execution.ask_fill.order, ask_entry.order)

    def test_less_buy_than_sell(self):
        bid_entry = OrderEntry(Order(Side.BUY, 'abc', 15, 130))
        ask_entry = OrderEntry(Order(Side.SELL, 'abc', 10, 110))

        execution = bid_entry.execute(ask_entry)

        self.assertEqual(execution.price, None)
        self.assertEqual(execution.quantity, 10)

        self.assertEqual(bid_entry.quantity, 15)
        self.assertEqual(ask_entry.quantity, 10)
        self.assertEqual(bid_entry.remaining, 5)
        self.assertEqual(ask_entry.remaining, 0)
        self.assertEqual(bid_entry.filled_quantity, 10)
        self.assertEqual(ask_entry.filled_quantity, 10)

        self.assertIs(execution.bid_fill.order, bid_entry.order)
        self.assertIs(execution.ask_fill.order, ask_entry.order)

if __name__ == '__main__':
    unittest.main()



from marketsim import KeyedHeap
import unittest

class TestKeyedHeap(unittest.TestCase):
    def test_heap(self):
        heap = KeyedHeap()
        self.assertFalse('key1' in heap)
        self.assertEqual(len(heap), 0)

        heap.push('key1', 'value1')
        self.assertTrue('key1' in heap)
        self.assertEqual(heap['key1'], 'value1')
        self.assertEqual(len(heap), 1)

        key, value = heap.peek()
        self.assertTrue('key1' in heap)
        self.assertEqual(key, 'key1')
        self.assertEqual(value, 'value1')
        self.assertEqual(heap['key1'], 'value1')
        self.assertEqual(len(heap), 1)

        key, value = heap.pop()
        self.assertFalse('key1' in heap)
        self.assertEqual(key, 'key1')
        self.assertEqual(value, 'value1')
        self.assertEqual(len(heap), 0)

    def test_priority(self):
        heap = KeyedHeap()
        heap.push('key2', 'bar')
        heap.push('key1', 'foo')
        heap.push('key3', 'baz')

        for expected_key, expected_value in [('key1', 'foo'), ('key2', 'bar'), ('key3', 'baz')]:
            key, value = heap.peek()
            self.assertEqual(key, expected_key)
            self.assertEqual(value, expected_value)

            key, value = heap.peek()
            self.assertEqual(key, expected_key)
            self.assertEqual(value, expected_value)

            key, value = heap.pop()
            self.assertEqual(key, expected_key)
            self.assertEqual(value, expected_value)

    def test_bool(self):
        heap = KeyedHeap()
        self.assertFalse(heap)
        self.assertFalse(heap.__bool__())
        self.assertFalse(heap.__nonzero__())
        self.assertTrue(heap.empty())

        heap.push('key1', 'value1')
        self.assertTrue(heap)
        self.assertTrue(heap.__bool__())
        self.assertTrue(heap.__nonzero__())
        self.assertFalse(heap.empty())

        heap.pop()
        self.assertFalse(heap)
        self.assertFalse(heap.__bool__())
        self.assertFalse(heap.__nonzero__())
        self.assertTrue(heap.empty())

    def test_exception(self):
        heap = KeyedHeap()

        with self.assertRaises(IndexError) as context:
            heap.peek()
        with self.assertRaises(IndexError) as context:
            heap.pop()

        heap.push('key1', 'value1')

        with self.assertRaises(KeyError) as context:
            heap.push('key1', 'value2')

        heap.pop()

        with self.assertRaises(IndexError) as context:
            heap.peek()
        with self.assertRaises(IndexError) as context:
            heap.pop()

    def test_iterate(self):
        heap = KeyedHeap()

        heap.push('key2', 'bar')
        heap.push('key1', 'foo')
        heap.push('key3', 'baz')

        self.assertEqual(list(heap.items()), [('key1', 'foo'), ('key2', 'bar'), ('key3', 'baz')])
        self.assertEqual(list(heap.keys()), ['key1', 'key2', 'key3'])
        self.assertEqual(list(heap.values()), ['foo', 'bar', 'baz'])

    def test_key_value(self):
        heap = KeyedHeap()

        heap.push('key1', 'value1')
        self.assertEqual(heap.peek_key(), 'key1')
        self.assertEqual(heap.pop_key(), 'key1')

        heap.push('key2', 'value2')
        self.assertEqual(heap.peek_value(), 'value2')
        self.assertEqual(heap.pop_value(), 'value2')



from heapq import heappush, heappop

class KeyedHeap:
    def __init__(self):
        self.pq_map = {}
        self.pq_list = []

    def __contains__(self, key):
        return key in self.pq_map

    def __getitem__(self, key):
        return self.pq_map[key]

    def __len__(self):
        return len(self.pq_list)

    def __bool__(self):
        return len(self.pq_list) != 0

    def __nonzero__(self):
        return len(self.pq_list) != 0

    def empty(self):
        return len(self.pq_list) == 0

    def push(self, key, value):
        if key in self:
            raise KeyError('key already exists: {}'.format(key))
        self.pq_map[key] = value
        heappush(self.pq_list, (key, value))

    def pop(self):
        if self.empty():
            raise IndexError('pop from an empty queue')
        key, value = heappop(self.pq_list)
        del self.pq_map[key]
        return (key, value)

    def pop_key(self):
        key, _ = self.pop()
        return key

    def pop_value(self):
        _, value = self.pop()
        return value

    def peek(self):
        if self.empty():
            raise IndexError('peek from an empty queue')
        key, value = self.pq_list[0]
        return (key, value)

    def peek_key(self):
        key, _ = self.peek()
        return key

    def peek_value(self):
        _, value = self.peek()
        return value

    def items(self):
        pq_list = list(self.pq_list)
        while len(pq_list) > 0:
            key, value = heappop(pq_list)
            yield key, value

    def keys(self):
        for key, _ in self.items():
            yield key

    def values(self):
        for _, value in self.items():
            yield value



from marketsim import Market, Order, Side, State, Product
import unittest

class TestMarket(unittest.TestCase):
    def execute(self, market, side, symbol, quantity, price = None):
        executions = market.execute(Order(side, symbol, quantity, price))
        return self.format_executions(executions)

    def format_executions(self, executions):
        return [(execution.quantity, execution.price) for execution in executions]

    def get_order_book(self, market, side, symbol):
        order_book = market[symbol][side].get_order_book()
        return [(entry.count, entry.volume, entry.price) for entry in order_book]

    def format_queue_stats(self, order_queue):
        return [
            order_queue.count,
            order_queue.volume,
            order_queue.market_order_count,
            order_queue.market_order_volume,
            order_queue.limit_order_count,
            order_queue.limit_order_volume,
        ]

    def test_market_products(self):
        market = Market()

        abcd = Product('abcd')
        self.assertIs(market.set_product('abcd', abcd), abcd)
        self.assertIs(market.get_product('abcd'), abcd)
        self.assertIs(market['abcd'], abcd)

        efgh = market.set_product('efgh', None)
        self.assertIsNotNone(market.get_product('efgh'))
        self.assertIsNotNone(market['efgh'])

        self.assertEqual(sorted(list(market)), ['abcd', 'efgh'])
        self.assertEqual(sorted(list(market.items())), [('abcd', abcd), ('efgh', efgh)])
        self.assertEqual(sorted(list(market.keys())), ['abcd', 'efgh'])
        self.assertEqual(sorted(list(market.values())), [abcd, efgh])

        self.assertEqual(set(market.get_products()), set([abcd, efgh]))

    def test_market_prices(self):
        market = Market()

        self.assertIsNone(market['abc'].bid_price)
        self.assertIsNone(market['abc'].ask_price)
        self.assertIsNone(market['abc'][Side.BUY].next_price)
        self.assertIsNone(market['abc'][Side.SELL].next_price)
        self.assertIsNone(market['abc'].last_price)

        market['abc'].last_price = 100
        self.assertEqual(market['abc'].last_price, 100)

        self.assertEqual(self.execute(market, Side.BUY, 'abc', 10, 100), [])
        self.assertEqual(self.execute(market, Side.BUY, 'abc', 10, 110), [])
        self.assertEqual(self.execute(market, Side.SELL, 'abc', 10, 120), [])
        self.assertEqual(self.execute(market, Side.SELL, 'abc', 10, 110), [(10, 110)])

        self.assertEqual(market['abc'].bid_price, 100)
        self.assertEqual(market['abc'].ask_price, 120)
        self.assertEqual(market['abc'][Side.BUY].next_price, 100)
        self.assertEqual(market['abc'][Side.SELL].next_price, 120)
        self.assertEqual(market['abc'].last_price, 110)

    def test_market_exceptions(self):
        market = Market()

        with self.assertRaises(KeyError):
            market.set_product(None, Product('invalid'))

        market.ensure_product('abc')

        with self.assertRaises(KeyError):
            market['abc']['invalid']

    def test_order_exceptions(self):
        market = Market()
        product = market.ensure_product('abc')
        order1 = Order(Side.BUY, 'abc', 10)

        # Order does not exist
        with self.assertRaises(ValueError):
            market.cancel(order1)
        with self.assertRaises(ValueError):
            product.cancel(order1)

        market.place(order1)

        # Order already exists
        with self.assertRaises(ValueError):
            market.place(order1)
        with self.assertRaises(ValueError):
            product.place(order1)

        order2 = Order(Side.SELL, 'abc', 20, 120)
        market.place(order2)
        market.execute()

        # Already fully filled  
        with self.assertRaises(ValueError):
            market.cancel(order1)

        market.cancel(order2)

        # Already cancelled
        with self.assertRaises(ValueError):
            market.cancel(order2)

    def test_limit_order_market_order(self):
        market = Market()
        self.assertEqual(self.execute(market, Side.BUY, 'abc', 10, 120), [])
        self.assertEqual(self.execute(market, Side.SELL, 'abc', 10, None), [(10, 120)])
        self.assertEqual(market['abc'].last_price, 120)
        self.assertIsNone(market['abc'].bid_price)
        self.assertIsNone(market['abc'].ask_price)

    def test_limit_order_limit_order(self):
        market = Market()
        self.assertEqual(self.execute(market, Side.BUY, 'abc', 10, 130), [])
        self.assertEqual(self.execute(market, Side.SELL, 'abc', 10, 110), [(10, 120)])
        self.assertEqual(market['abc'].last_price, 120)
        self.assertIsNone(market['abc'].bid_price)
        self.assertIsNone(market['abc'].ask_price)

    def test_market_order_limit_order(self):
        market = Market()
        self.assertEqual(self.execute(market, Side.BUY, 'abc', 10, None), [])
        self.assertEqual(self.execute(market, Side.SELL, 'abc', 10, 120), [(10, 120)])
        self.assertEqual(market['abc'].last_price, 120)
        self.assertIsNone(market['abc'].bid_price)
        self.assertIsNone(market['abc'].ask_price)

    def test_market_order_market_order(self):
        market = Market()
        self.assertEqual(self.execute(market, Side.BUY, 'abc', 10, None), [])
        self.assertEqual(self.execute(market, Side.SELL, 'abc', 10, None), [])
        self.assertEqual(market['abc'].last_price, None)
        self.assertIsNone(market['abc'].bid_price)
        self.assertIsNone(market['abc'].ask_price)

    def test_trading(self):
        market = Market()
        self.assertEqual(self.execute(market, Side.SELL, 'abc', 40, 130), [])
        self.assertEqual(self.execute(market, Side.SELL, 'abc', 80, 130), [])
        self.assertEqual(self.execute(market, Side.SELL, 'abc', 10, 120), [])
        self.assertEqual(self.execute(market, Side.SELL, 'abc', 20, 120), [])

        self.assertEqual(self.get_order_book(market, Side.BUY, 'abc'), [])
        self.assertEqual(self.get_order_book(market, Side.SELL, 'abc'), [(2, 30, 120), (2, 120, 130)])

        self.assertEqual(self.execute(market, Side.BUY, 'abc', 45, None), [(10, 130), (20, 130), (15, 130)])
        self.assertEqual(market['abc'].last_price, 130)
        self.assertIsNone(market['abc'].bid_price)
        self.assertEqual(market['abc'].ask_price, 130)

        self.assertEqual(self.get_order_book(market, Side.BUY, 'abc'), [])
        self.assertEqual(self.get_order_book(market, Side.SELL, 'abc'), [(2, 105, 130)])

    def test_auction(self):
        market = Market()
        market.place(Order(Side.SELL, 'abc', 40, 130, time=0))
        market.place(Order(Side.SELL, 'abc', 80, 130, time=0))
        market.place(Order(Side.SELL, 'abc', 10, 120, time=0))
        market.place(Order(Side.SELL, 'abc', 20, 120, time=0))
        market.place(Order(Side.BUY, 'abc', 45, None, time=0))

        self.assertEqual(self.get_order_book(market, Side.BUY, 'abc'), [(1, 45, None)])
        self.assertEqual(self.get_order_book(market, Side.SELL, 'abc'), [(2, 30, 120), (2, 120, 130)])

        self.assertEqual(self.format_executions(market.execute()), [(10, 130), (20, 130), (5, 130), (10, 130)])
        self.assertEqual(market['abc'].last_price, 130)
        self.assertIsNone(market['abc'].bid_price)
        self.assertEqual(market['abc'].ask_price, 130)

        self.assertEqual(self.get_order_book(market, Side.BUY, 'abc'), [])
        self.assertEqual(self.get_order_book(market, Side.SELL, 'abc'), [(2, 105, 130)])

    def test_auction_with_remaining_market_orders(self):
        market = Market()
        market.place(Order(Side.SELL, 'abc', 40, 130, time=0))
        market.place(Order(Side.SELL, 'abc', 80, 130, time=0))
        market.place(Order(Side.SELL, 'abc', 10, 120, time=0))
        market.place(Order(Side.SELL, 'abc', 20, 120, time=0))
        market.place(Order(Side.SELL, 'abc', 60, None, time=0))
        market.place(Order(Side.BUY, 'abc', 45, 110, time=0))

        self.assertEqual(self.get_order_book(market, Side.BUY, 'abc'), [(1, 45, 110)])
        self.assertEqual(self.get_order_book(market, Side.SELL, 'abc'), [(1, 60, None), (2, 30, 120), (2, 120, 130)])

        self.assertEqual(self.format_executions(market.execute()), [(45, 110)])
        self.assertEqual(market['abc'].last_price, 110)
        self.assertIsNone(market['abc'].bid_price)
        self.assertEqual(market['abc'].ask_price, 120)

        self.assertEqual(self.get_order_book(market, Side.BUY, 'abc'), [])
        self.assertEqual(self.get_order_book(market, Side.SELL, 'abc'), [(1, 15, None), (2, 30, 120), (2, 120, 130)])

    def test_auction_cannot_execute_market_orders_only(self):
        market = Market()
        market.place(Order(Side.SELL, 'abc', 10, None, time=0))
        market.place(Order(Side.BUY, 'abc', 10, None, time=0))
        self.assertEqual(market.execute(), [])

    def test_auction_cannot_execute_spread_limit_orders(self):
        market = Market()
        market.place(Order(Side.SELL, 'abc', 10, 130, time=0))
        market.place(Order(Side.BUY, 'abc', 10, 110, time=0))
        market.place(Order(Side.SELL, 'abc', 10, None, time=0))
        market.place(Order(Side.BUY, 'abc', 10, None, time=0))
        self.assertEqual(market.execute(), [])

    def test_auction_cannot_execute_insufficient_bid_limit_order(self):
        market = Market()
        market.place(Order(Side.SELL, 'abc', 10, 130, time=0))
        market.place(Order(Side.SELL, 'abc', 15, None, time=0))
        market.place(Order(Side.BUY, 'abc', 10, None, time=0))
        self.assertEqual(market.execute(), [])

    def test_auction_cannot_execute_insufficient_ask_limit_order(self):
        market = Market()
        market.place(Order(Side.BUY, 'abc', 10, 110, time=0))
        market.place(Order(Side.SELL, 'abc', 10, None, time=0))
        market.place(Order(Side.BUY, 'abc', 15, None, time=0))
        self.assertEqual(market.execute(), [])

    def test_cancel(self):
        market = Market()
        order1 = Order(Side.BUY, 'abc', 10, 120)
        order2 = Order(Side.BUY, 'abc', 10, 110)
        market.place(order1)
        market.place(order2)
        market.cancel(order1)
        market.place(Order(Side.SELL, 'abc', 10))
        self.assertEqual(self.format_executions(market.execute()), [(10, 110)])

    def test_cancel_by_id(self):
        market = Market()
        order1 = Order(Side.BUY, 'abc', 10, 120, id='order1')
        order2 = Order(Side.BUY, 'abc', 10, 110, id='order2')
        market.place(order1)
        market.place(order2)
        market.cancel(Order(id='order1'))
        market.place(Order(Side.SELL, 'abc', 10))
        self.assertEqual(self.format_executions(market.execute()), [(10, 110)])

    def test_execute(self):
        market = Market()
        order1 = Order(Side.BUY, 'abc', 10, 120, time=1, id='order1')
        order2 = Order(Side.SELL, 'abc', 20, None, time=2, id='order2')
        self.assertEqual(market.execute(order1), [])
        executions = market.execute(order2)
        self.assertEqual(self.format_executions(executions), [(10, 120)])
        fill1 = executions[0].fills[Side.BUY]
        fill2 = executions[0].fills[Side.SELL]

        self.assertEqual(fill1.order, order1)
        self.assertEqual(fill1.quantity, 10)
        self.assertEqual(fill1.price, 120)
        self.assertEqual(fill1.side, Side.BUY)
        self.assertEqual(fill1.symbol, 'abc')
        self.assertEqual(fill1.order_quantity, 10)
        self.assertEqual(fill1.order_price, 120)
        self.assertEqual(fill1.order_time, 1)
        self.assertEqual(fill1.order_id, 'order1')
        self.assertEqual(fill1.cumulative_quantity, 10)

        self.assertEqual(fill2.order, order2)
        self.assertEqual(fill2.quantity, 10)
        self.assertEqual(fill2.price, 120)
        self.assertEqual(fill2.side, Side.SELL)
        self.assertEqual(fill2.symbol, 'abc')
        self.assertEqual(fill2.order_quantity, 20)
        self.assertEqual(fill2.order_price, None)
        self.assertEqual(fill2.order_time, 2)
        self.assertEqual(fill2.order_id, 'order2')
        self.assertEqual(fill2.cumulative_quantity, 10)

    def test_execute_after_cancel(self):
        market = Market()
        order = Order(Side.BUY, 'abc', 100, 110)
        market.place(order)
        market.cancel(order)
        self.assertEqual(market.execute(), [])

    def test_market_shortcuts(self):
        market = Market()

        market.place_order('buy', 'abc', 10, 110, id='order1')
        self.assertEqual(market.entries['order1'].state, State.NEW)

        market.cancel_order(id='order1')
        self.assertEqual(market.entries['order1'].state, State.CANCELLED)

        market.execute_order('buy', 'abc', 10, 120, id='order2')
        market.execute_order('sell', 'abc', 20, 120, id='order3')
        self.assertEqual(market.entries['order2'].state, State.FULLY_FILLED)
        self.assertEqual(market.entries['order3'].state, State.PARTIALLY_FILLED)

    def test_product_shortcuts(self):
        market = Market()
        product = market.ensure_product('abc')

        product.place_order('buy', 'abc', 10, 110, id='order1')
        self.assertEqual(product.entries['order1'].state, State.NEW)

        product.cancel_order(id='order1')
        self.assertEqual(product.entries['order1'].state, State.CANCELLED)

        product.execute_order('buy', 'abc', 10, 120, id='order2')
        product.execute_order('sell', 'abc', 20, 120, id='order3')
        self.assertEqual(product.entries['order2'].state, State.FULLY_FILLED)
        self.assertEqual(product.entries['order3'].state, State.PARTIALLY_FILLED)

    def test_order_queue_stats(self):
        market = Market()

        sell1 = Order(Side.SELL, 'abc', 10, 130, id='sell1')
        sell2 = Order(Side.SELL, 'abc', 10, 120, id='sell2')
        sell3 = Order(Side.SELL, 'abc', 10, 110, id='sell3')
        sell4 = Order(Side.SELL, 'abc', 10, None, id='sell4')

        buy1 = Order(Side.BUY, 'abc', 10, 110, id='buy1')
        buy2 = Order(Side.BUY, 'abc', 10, 100, id='buy2')
        buy3 = Order(Side.BUY, 'abc', 10, 90, id='buy3')
        buy4 = Order(Side.BUY, 'abc', 10, None, id='buy4')

        for order in (sell1, sell2, sell3, sell4, buy1, buy2, buy3, buy4):
            market.place(order)

        self.assertEqual(self.format_queue_stats(market['abc'][Side.BUY]), [4, 40, 1, 10, 3, 30])
        self.assertEqual(self.format_queue_stats(market['abc'][Side.SELL]), [4, 40, 1, 10, 3, 30])

        market.execute()
        self.assertEqual(self.format_queue_stats(market['abc'][Side.BUY]), [2, 20, 0, 0, 2, 20])
        self.assertEqual(self.format_queue_stats(market['abc'][Side.SELL]), [2, 20, 0, 0, 2, 20])

        market.cancel(buy3)
        self.assertEqual(self.format_queue_stats(market['abc'][Side.BUY]), [1, 10, 0, 0, 1, 10])
        self.assertEqual(self.format_queue_stats(market['abc'][Side.SELL]), [2, 20, 0, 0, 2, 20])

        market.cancel(sell1)
        self.assertEqual(self.format_queue_stats(market['abc'][Side.BUY]), [1, 10, 0, 0, 1, 10])
        self.assertEqual(self.format_queue_stats(market['abc'][Side.SELL]), [1, 10, 0, 0, 1, 10])

    def test_order_queue_internal(self):
        market = Market()

        sell1 = Order(Side.SELL, 'abc', 10, 130, time=1, id='sell1') # filled
        sell2 = Order(Side.SELL, 'abc', 10, 120, time=2, id='sell2') # top of price & time queues
        sell3 = Order(Side.SELL, 'abc', 10, 120, time=3, id='sell3') # top of price queue
        sell4 = Order(Side.SELL, 'abc', 10, 110, time=4, id='sell4')
        sell5 = Order(Side.SELL, 'abc', 10, None, time=5, id='sell5') # filled

        buy1 = Order(Side.BUY, 'abc', 10, 110, time=6, id='buy1') # filled
        buy2 = Order(Side.BUY, 'abc', 10, 100, time=7, id='buy2') # top of price & time queue
        buy3 = Order(Side.BUY, 'abc', 10, 100, time=8, id='buy3') # top of price queue
        buy4 = Order(Side.BUY, 'abc', 10, 90, time=9, id='buy4')
        buy5 = Order(Side.BUY, 'abc', 10, None, time=10, id='buy5') # filled

        for order in (sell1, sell2, sell3, sell4, sell5, buy1, buy2, buy3, buy4, buy5):
            market.place(order)

        market.execute()

        buy_order_queue = market['abc'][Side.BUY]
        sell_order_queue = market['abc'][Side.SELL]
        self.assertEqual(buy_order_queue.next_price, 100)
        self.assertEqual(sell_order_queue.next_price, 120)

        buy_price_queue = buy_order_queue.heap.peek_value()
        sell_price_queue = sell_order_queue.heap.peek_value()
        self.assertEqual(buy_price_queue.price, 100)
        self.assertEqual(buy_price_queue.count, 2)
        self.assertEqual(buy_price_queue.volume, 20)
        self.assertEqual(sell_price_queue.price, 120)
        self.assertEqual(sell_price_queue.count, 2)
        self.assertEqual(sell_price_queue.volume, 20)

        buy_time_queue = buy_price_queue.heap.peek_value()
        sell_time_queue = sell_price_queue.heap.peek_value()
        self.assertEqual(buy_time_queue.time, 7)
        self.assertEqual(buy_time_queue.count, 1)
        self.assertEqual(buy_time_queue.volume, 10)
        self.assertEqual(sell_time_queue.time, 2)
        self.assertEqual(sell_time_queue.count, 1)
        self.assertEqual(sell_time_queue.volume, 10)

    def test_order_persistence(self):
        market = Market()
        self.assertIs(market.get_order_by_id('order1'), None)
        self.assertIs(market.get_order_by_id('order2'), None)
        self.assertIs(market['abc'].get_order_by_id('order1'), None)
        self.assertIs(market['abc'].get_order_by_id('order2'), None)

        order1 = Order(Side.BUY, 'abc', 10, 90, id='order1')
        market.place(order1)
        self.assertIs(market.get_order_by_id('order1'), order1)
        self.assertIs(market.get_order_by_id('order2'), None)
        self.assertIs(market['abc'].get_order_by_id('order1'), order1)
        self.assertIs(market['abc'].get_order_by_id('order2'), None)

        order2 = Order(Side.SELL, 'abc', 20, None, id='order2')
        market.execute(order2)
        self.assertIs(market.get_order_by_id('order1'), order1)
        self.assertIs(market.get_order_by_id('order2'), order2)
        self.assertIs(market['abc'].get_order_by_id('order1'), order1)
        self.assertIs(market['abc'].get_order_by_id('order2'), order2)

        market.cancel(order2)
        self.assertIs(market.get_order_by_id('order1'), order1)
        self.assertIs(market.get_order_by_id('order2'), order2)
        self.assertIs(market['abc'].get_order_by_id('order1'), order1)
        self.assertIs(market['abc'].get_order_by_id('order2'), order2)

if __name__ == '__main__':
    unittest.main()



from marketsim import Order, OrderEntry, Side, TimeOrderQueue
import unittest

class TestAllocation(unittest.TestCase):
    def allocate(self, queue, sum_quantity):
        allocations = queue.allocate(sum_quantity)
        return [(alloc.quantity, alloc.entry.remaining, alloc.entry.quantity) for alloc in allocations]

    def test_allocation(self):
        bid_queue = TimeOrderQueue(None)
        bid_queue.push(OrderEntry(Order(Side.BUY, 'abc', 10, 120)))
        bid_queue.push(OrderEntry(Order(Side.BUY, 'abc', 20, 120)))
        bid_queue.push(OrderEntry(Order(Side.BUY, 'abc', 30, 120)))

        self.assertEqual(self.allocate(bid_queue, 18), [(3, 10, 10), (6, 20, 20), (9, 30, 30)])

        ask_queue = TimeOrderQueue(None)
        ask_queue.push(OrderEntry(Order(Side.SELL, 'abc', 18, None)))
        bid_queue.execute(ask_queue)

        TimeOrderQueue.DEBUG = True
        self.assertEqual(self.allocate(bid_queue, 35), [(6, 7, 10), (12, 14, 20), (17, 21, 30)])
        TimeOrderQueue.DEBUG = False

    def test_adjust_allocation(self):
        bid_queue = TimeOrderQueue(None)
        bid_queue.push(OrderEntry(Order(Side.BUY, 'abc', 11, 120)))
        bid_queue.push(OrderEntry(Order(Side.BUY, 'abc', 13, 120)))
        bid_queue.push(OrderEntry(Order(Side.BUY, 'abc', 17, 120)))
        bid_queue.push(OrderEntry(Order(Side.BUY, 'abc', 19, 120)))
        bid_queue.push(OrderEntry(Order(Side.BUY, 'abc', 23, 120)))
        # 11 + 13 + 17 + 19 + 23 == 83
        # 11 / 83 == 0.13
        # 13 / 83 == 0.15
        # 17 / 83 == 0.20
        # 19 / 83 == 0.22
        # 23 / 83 == 0.27

        # Allocate 41
        # [5.43, 6.42, 8.39, 9.38, 11.36]
        # [5, 6, 8, 9, 11] #=> sum == 39
        # [5+1, 6+1, 8, 9, 11]
        self.assertEqual(self.allocate(bid_queue, 41), [(6, 11, 11), (7, 13, 13), (8, 17, 17), (9, 19, 19), (11, 23, 23)])

        # Allocate 42
        # [5.56, 6.57, 8.60, 9.61, 11.63]
        # [6, 7, 9, 10, 12] #=> sum == 44
        # [6, 7, 9, 10-1, 12-1]
        self.assertEqual(self.allocate(bid_queue, 42), [(6, 11, 11), (7, 13, 13), (9, 17, 17), (9, 19, 19), (11, 23, 23)])

if __name__ == '__main__':
    unittest.main()



from collections import deque
from datetime import datetime
from enum import Enum
from time import mktime
from marketsim.keyed_heap import KeyedHeap

builtin_id = id

class Side(Enum):
    BUY  = 1
    SELL = 2

    @classmethod
    def normalize(cls, value):
        if value is None:
            return None
        elif isinstance(value, cls):
            return value
        elif isinstance(value, str):
            return cls[value.upper()]
        elif isinstance(value, int):
            return cls(value)
        else:
            raise ValueError('invalid side: {}'.format(value))

class State(Enum):
    NEW              = 0
    PARTIALLY_FILLED = 1
    FULLY_FILLED     = 2
    CANCELLED        = 3

class Order:
    def __init__(self, side=None, symbol=None, quantity=None, price=None, time=None, id=None):
        self._side = Side.normalize(side)
        self._symbol = symbol
        self._quantity = quantity
        self._price = price
        self._time = time
        self._id = id if id is not None else builtin_id(self)

    @property
    def side(self):
        return self._side

    @property
    def symbol(self):
        return self._symbol

    @property
    def quantity(self):
        return self._quantity

    @property
    def price(self):
        return self._price

    @property
    def time(self):
        return self._time

    @property
    def id(self):
        return self._id

    def __repr__(self):
        return 'Order(side={}, symbol={}, quantity={}, price={})'.format(self.side, self.symbol, self.quantity, self.price)

class Fill:
    def __init__(self, entry, quantity, price=None):
        self._order = entry.order
        self._quantity = quantity
        self._price = price if price is not None else entry.price

        self._side = entry.side
        self._symbol = entry.symbol
        self._order_quantity = entry.quantity
        self._order_price = entry.price
        self._order_time = entry.time
        self._order_id = entry.order_id
        self._cumulative_quantity = entry.filled_quantity

    @property
    def order(self):
        return self._order

    @property
    def quantity(self):
        return self._quantity

    @property
    def price(self):
        return self._price

    @property
    def side(self):
        return self._side

    @property
    def symbol(self):
        return self._symbol

    @property
    def order_quantity(self):
        return self._order_quantity

    @property
    def order_price(self):
        return self._order_price

    @property
    def order_time(self):
        return self._order_time

    @property
    def order_id(self):
        return self._order_id

    @property
    def cumulative_quantity(self):
        return self._cumulative_quantity

    def __repr__(self):
        return "Fill(side={}, symbol={}, quantity={}, price={}, cumulative_quantity={})".format(self.side, self.symbol, self.quantity, self.price, self.cumulative_quantity)

class Execution:
    def __init__(self, bid_entry, ask_entry, quantity, price=None):
        self._quantity = quantity
        self._price = price

        self._fills = {
            Side.BUY : Fill(bid_entry, quantity, price),
            Side.SELL: Fill(ask_entry, quantity, price),
        }

    @property
    def quantity(self):
        return self._quantity

    @property
    def price(self):
        return self._price

    @property
    def fills(self):
        return self._fills

    @property
    def bid_fill(self):
        return self.fills[Side.BUY]

    @property
    def ask_fill(self):
        return self.fills[Side.SELL]

    def __repr__(self):
        return "Execution(bid_fill={}, ask_fill={}, quantity={}, price={})".format(self.bid_fill, self.ask_fill, self.quantity, self.price)

class OrderEntry:
    def __init__(self, order):
        self._order = order

        self._side = order.side
        self._symbol = order.symbol
        self._quantity = order.quantity
        self._price = order.price
        self._time = order.time if order.time is not None else self.default_time()
        self._order_id = order.id

        self._remaining = order.quantity
        self._state = State.NEW

    @property
    def order(self):
        return self._order

    @property
    def side(self):
        return self._side

    @property
    def symbol(self):
        return self._symbol

    @property
    def quantity(self):
        return self._quantity

    @property
    def price(self):
        return self._price

    @property
    def time(self):
        return self._time

    @property
    def order_id(self):
        return self._order_id

    @property
    def remaining(self):
        return self._remaining

    @property
    def state(self):
        return self._state

    @property
    def filled_quantity(self):
        return self.quantity - self.remaining

    def default_time(self):
        now = datetime.now()
        return mktime(now.timetuple()) + now.microsecond / 1000000.0

    def cancel(self):
        self._remaining = 0
        self._state = State.CANCELLED

    def execute(self, ask_entry, quantity=None):
        bid_entry = self

        if quantity is None:
            quantity = min(bid_entry.remaining, ask_entry.remaining)

        bid_entry._remaining -= quantity
        ask_entry._remaining -= quantity

        bid_entry._state = State.FULLY_FILLED if bid_entry.remaining == 0 else State.PARTIALLY_FILLED
        ask_entry._state = State.FULLY_FILLED if ask_entry.remaining == 0 else State.PARTIALLY_FILLED

        return Execution(bid_entry, ask_entry, quantity)

class OrderStat:
    def __init__(self, price, volume, count):
        self._price  = price
        self._volume = volume
        self._count  = count

    @property
    def price(self):
        return self._price

    @property
    def volume(self):
        return self._volume

    @property
    def count(self):
        return self._count

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return 'OrderStat(price={}, volume={}, count={})'.format(self.price, self.volume, self.count)

class Allocation:
    def __init__(self, entry, quantity):
        self._entry = entry
        self._quantity = quantity

    @property
    def entry(self):
        return self._entry

    @property
    def quantity(self):
        return self._quantity

    def __repr__(self):
        return 'Allocation(entry={}, quantity={})'.format(self.entry, self.quantity)

    @property
    def pair(self):
        return (self.entry, self.quantity)

class TimeOrderQueue:
    def __init__(self, time):
        self._time = time
        self._volume = 0
        self._entries = deque()

    @property
    def time(self):
        return self._time

    @property
    def volume(self):
        return self._volume

    @property
    def entries(self):
        return self._entries

    @property
    def count(self):
        return len(self.entries)

    def empty(self):
        return len(self.entries) == 0

    def push(self, entry):
        self._volume += entry.remaining
        self.entries.append(entry)
        return self

    def cancel(self, entry):
        self._volume -= entry.remaining
        entry.cancel()
        return self

    def execute(self, ask_queue):
        bid_queue = self

        sum_quantity = min(bid_queue.volume, ask_queue.volume)

        if sum_quantity == 0:
            return []

        bid_allocations = bid_queue.allocate(sum_quantity)
        ask_allocations = ask_queue.allocate(sum_quantity)

        executions = []
        b = 0
        a = 0
        bid_entry, bid_quantity = bid_allocations[0].pair
        ask_entry, ask_quantity = ask_allocations[0].pair

        while b < len(bid_allocations) and a < len(ask_allocations):
            quantity = min(bid_quantity, ask_quantity)
            execution = bid_entry.execute(ask_entry, quantity)
            executions.append(execution)

            bid_queue._volume -= execution.quantity
            ask_queue._volume -= execution.quantity
            bid_quantity -= execution.quantity
            ask_quantity -= execution.quantity

            if bid_quantity == 0:
                b += 1
                if b < len(bid_allocations):
                    bid_entry, bid_quantity = bid_allocations[b].pair
            if ask_quantity == 0:
                a += 1
                if a < len(ask_allocations):
                    ask_entry, ask_quantity = ask_allocations[a].pair

        return executions

    def allocate(self, sum_quantity):
        unit = float(sum_quantity) / self.volume
        remaining_entries = filter(lambda entry: entry.remaining > 0, self.entries)
        allocations = [Allocation(entry, round(entry.remaining * unit)) for entry in remaining_entries]

        current_sum = sum([allocation.quantity for allocation in allocations])

        while current_sum != sum_quantity:
            if current_sum < sum_quantity:
                incr = 1
                alloc_iter = iter(allocations)
            else:
                incr = -1
                alloc_iter = reversed(allocations)

            while current_sum != sum_quantity:
                allocation = next(alloc_iter)
                allocation._quantity += incr
                current_sum += incr

        return allocations

class PriceOrderQueue:
    def __init__(self, price):
        self._heap = KeyedHeap()
        self._price = price
        self._count = 0
        self._volume = 0

    @property
    def heap(self):
        return self._heap

    @property
    def price(self):
        return self._price

    @property
    def count(self):
        return self._count

    @property
    def volume(self):
        return self._volume

    def empty(self):
        return self.heap.empty()

    def get_time_key(self, entry):
        return entry.time

    def push(self, entry):
        time = entry.time
        time_key = self.get_time_key(entry)

        if time_key in self.heap:
            child = self.heap[time_key]
        else:
            child = TimeOrderQueue(time)
            self.heap.push(time_key, child)

        child.push(entry)
        self._count += 1
        self._volume += entry.remaining

        return self

    def cancel(self, entry):
        time_key = self.get_time_key(entry)
        assert time_key in self.heap, 'order entry does not exist for time_key: {}'.format(time_key)

        # Note: Update volume before child.cancel(entry). Otherwise, entry.remaining would already be zero.
        self._count -= 1
        self._volume -= entry.remaining
        child = self.heap[time_key]
        child.cancel(entry)

        return self

    def pop_empty_values(self):
        while not self.heap.empty():
            if self.heap.peek_value().empty() or self.heap.peek_value().volume == 0:
                self.heap.pop()
            else:
                break

    def execute(self, ask_queue):
        bid_queue = self

        executions = []

        while not bid_queue.heap.empty() and not ask_queue.heap.empty():
            bid_child = bid_queue.heap.peek_value()
            ask_child = ask_queue.heap.peek_value()

            bid_orig_count = bid_child.count
            ask_orig_count = ask_child.count

            child_executions = bid_child.execute(ask_child)

            bid_queue._count -= bid_orig_count - bid_child.count
            ask_queue._count -= ask_orig_count - ask_child.count

            bid_queue.pop_empty_values()
            ask_queue.pop_empty_values()

            for execution in child_executions:
                bid_queue._volume -= execution.quantity
                ask_queue._volume -= execution.quantity

            executions.extend(child_executions)

        return executions

class OrderQueue:
    def __init__(self):
        self._heap = KeyedHeap()
        self._count = 0
        self._volume = 0
        self._market_order_count = 0
        self._market_order_volume = 0
        self._limit_order_count = 0
        self._limit_order_volume = 0
        self._next_price = None

    @property
    def heap(self):
        return self._heap

    @property
    def count(self):
        return self._count

    @property
    def volume(self):
        return self._volume

    @property
    def market_order_count(self):
        return self._market_order_count

    @property
    def market_order_volume(self):
        return self._market_order_volume

    @property
    def limit_order_count(self):
        return self._limit_order_count

    @property
    def limit_order_volume(self):
        return self._limit_order_volume

    @property
    def next_price(self):
        return self._next_price

    def update_stats(self, delta_count, delta_quantity, is_market_order):
        self._count += delta_count
        self._volume += delta_quantity
        if is_market_order:
            self._market_order_count += delta_count
            self._market_order_volume += delta_quantity
        else:
            self._limit_order_count += delta_count
            self._limit_order_volume += delta_quantity

    def update_next_price(self):
        if self.heap.empty():
            self._next_price = None
        elif self.heap.peek_value().price is not None:
            self._next_price = self.heap.peek_value().price
        else:
            # Put aside market order temporarily to find the next limit price
            key, child = self.heap.pop()
            if self.heap.empty():
                self._next_price = None
            else:
                self._next_price = self.heap.peek_value().price
            self.heap.push(key, child)

    def get_price_key(self, entry):
        price = entry.price
        side = entry.side

        if price is None:
            price_key = float('-inf')
        elif side == Side.BUY:
            price_key = -price
        else:
            price_key = price

        return price_key

    def pop_empty_values(self):
        while not self.heap.empty():
            if self.heap.peek_value().empty() or self.heap.peek_value().volume == 0:
                self.heap.pop()
            else:
                break

    def push(self, entry):
        price = entry.price
        price_key = self.get_price_key(entry)

        if price_key in self.heap:
            child = self.heap[price_key]
        else:
            child = PriceOrderQueue(price)
            self.heap.push(price_key, child)

        child.push(entry)
        self.update_stats(+1, entry.remaining, entry.price is None)
        self.update_next_price()

        return self

    def cancel(self, entry):
        price_key = self.get_price_key(entry)
        assert price_key in self.heap, 'order entry does not exist for price_key: {}'.format(price_key)

        # Note: Update stats before child.cancel(entry). Otherwise, entry.remaining would already be zero.
        self.update_stats(-1, -entry.remaining, entry.price is None)
        self.update_next_price()
        child = self.heap[price_key]
        child.cancel(entry)

        return self

    def can_execute(self, ask_queue):
        """
        In order for a product to be executable, certain conditions must be
        satisfied to be able to determine the execution price, depending on the
        currently queued orders.

        For example, the execution price cannot be determined solely by market-
        orders from both bid and ask sides, without limit-orders somehow
        involved.

        Case 1: Equal market-order volume on both bid and ask sides:
            There must be limit-orders on both sides. In addition, the limit-
            orders must include executable prices that cross each other.

        Case 2: There are more market-orders on bid side:
            After all market-orders on ask side are executed, there must be some
            limit-orders on ask side, in order to execute the remaining market-
            orders on bid side.

        Case 3: There are more market-orders on ask side:
            Same as above, but in the opposite way.
        """
        bid_queue = self

        if bid_queue.market_order_volume == ask_queue.market_order_volume:
            if bid_queue.limit_order_volume == 0 or ask_queue.limit_order_volume == 0:
                return False
            elif bid_queue.next_price < ask_queue.next_price:
                return False
        elif bid_queue.market_order_volume < ask_queue.market_order_volume:
            if bid_queue.limit_order_volume == 0:
                return False
        else:
            if ask_queue.limit_order_volume == 0:
                return False

        return True

    def execute(self, ask_queue):
        bid_queue = self

        if not bid_queue.can_execute(ask_queue):
            return []

        executions = []

        while not bid_queue.heap.empty() and not ask_queue.heap.empty():
            bid_child = bid_queue.heap.peek_value()
            ask_child = ask_queue.heap.peek_value()

            if bid_child.price is not None and ask_child.price is not None:
                if bid_child.price < ask_child.price:
                    break

            child_executions = bid_child.execute(ask_child)

            bid_queue.pop_empty_values()
            ask_queue.pop_empty_values()

            executions.extend(child_executions)

        assert executions, 'there were no executions. may be a bug in can_execute()'

        bid_price = executions[-1].bid_fill.price
        ask_price = executions[-1].ask_fill.price

        if bid_price is None:
            price = ask_price
        elif ask_price is None:
            price = bid_price
        else:
            price = (bid_price + ask_price) / 2

        assert price is not None

        for execution in executions:
            bid_delta = -1 if execution.bid_fill.cumulative_quantity == execution.bid_fill.order_quantity else 0
            ask_delta = -1 if execution.ask_fill.cumulative_quantity == execution.ask_fill.order_quantity else 0
            bid_queue.update_stats(bid_delta, -execution.quantity, execution.bid_fill.price is None)
            ask_queue.update_stats(ask_delta, -execution.quantity, execution.ask_fill.price is None)
            execution._price = price
            execution.bid_fill._price = price
            execution.ask_fill._price = price

        bid_queue.update_next_price()
        ask_queue.update_next_price()

        return executions

    def get_order_book(self):
        order_book = []
        for child in self.heap.values():
            if child.volume > 0:
                order_book.append(OrderStat(child.price, child.volume, child.count))
        return order_book

class Product:
    def __init__(self, symbol):
        self._symbol = symbol

        self._order_queues = {
            Side.BUY : OrderQueue(),
            Side.SELL: OrderQueue(),
        }

        self._entries = {}
        self._last_price = None

    @property
    def symbol(self):
        return self._symbol

    @property
    def order_queues(self):
        return self._order_queues

    @property
    def entries(self):
        return self._entries

    @property
    def bid_price(self):
        return self.order_queues[Side.BUY].next_price

    @property
    def ask_price(self):
        return self.order_queues[Side.SELL].next_price

    @property
    def last_price(self):
        return self._last_price

    @last_price.setter
    def last_price(self, last_price):
        self._last_price = last_price

    def __lt__(self, other):
        return self.symbol < other.symbol

    def __getitem__(self, side):
        side = Side.normalize(side)
        return self.order_queues[side]

    def place(self, order):
        if order.id in self.entries:
            raise ValueError('duplicate order id')

        entry = OrderEntry(order)
        self.order_queues[order.side].push(entry)
        self.entries[order.id] = entry

    def cancel(self, order):
        if order.id not in self.entries:
            raise ValueError('no such order id')

        # Note: Update volume before queue.cancel(entry). Otherwise, entry.remaining would already be zero.
        entry = self.entries[order.id]

        if entry.state == State.FULLY_FILLED:
            raise ValueError('already fully filled')
        if entry.state == State.CANCELLED:
            raise ValueError('already cancelled')

        self.order_queues[entry.side].cancel(entry)

    def execute(self, order=None):
        if order is not None:
            self.place(order)

        bid_order_queue = self.order_queues[Side.BUY]
        ask_order_queue = self.order_queues[Side.SELL]

        executions = bid_order_queue.execute(ask_order_queue)

        if executions:
            self._last_price = executions[-1].price

        return executions

    def place_order(self, *args, **kwargs):
        return self.place(Order(*args, **kwargs))

    def cancel_order(self, *args, **kwargs):
        return self.cancel(Order(*args, **kwargs))

    def execute_order(self, *args, **kwargs):
        return self.execute(Order(*args, **kwargs))

    def get_order_by_id(self, order_id):
        if order_id not in self.entries:
            return None
        entry = self.entries[order_id]
        return entry.order

    def format_order_book(self):
        bid_order_book = self.order_queues[Side.BUY].get_order_book()
        ask_order_book = self.order_queues[Side.SELL].get_order_book()

        entries = {}

        for side, order_book in [(Side.BUY, bid_order_book), (Side.SELL, ask_order_book)]:
            for entry in order_book:
                if entry.price not in entries:
                    entries[entry.price] = {}
                entries[entry.price][side] = entry

        rows = []
        rows.append(['BID', 'PRICE', 'ASK'])
        rows.append(['===', '=====', '==='])

        for price in reversed(sorted(list(entries.keys()))):
            values = {Side.BUY: '', Side.SELL: ''}
            for side in [Side.BUY, Side.SELL]:
                if side in entries[price]:
                    entry = entries[price][side]
                    values[side] = '{} ({})'.format(entry.volume, entry.count)
            rows.append([values[Side.BUY], str(price), values[Side.SELL]])

        max_len = [max(len(row[i]) for row in rows) for i in range(3)]

        result = []

        for row in rows:
            values = [('{:' + str(max_len[i]) + '}').format(row[i]) for i in range(3)]
            result.append('| ' + ' | '.join(values) + ' |')

        result[1] = '|' + '|'.join(['=' * (max_len[i] + 2) for i in range(3)]) + '|'

        return "\n".join(result)

class Market:
    def __init__(self):
        self._products = {}
        self._entries = {}

    @property
    def products(self):
        return self._products

    @property
    def entries(self):
        return self._entries

    def __contains__(self, symbol):
        return symbol in self.products

    def __setitem__(self, symbol, product):
        if symbol is None:
            raise KeyError('symbol must be specified')
        if product is None:
            product = Product(symbol)
        self.products[symbol] = product
        return product

    def __getitem__(self, symbol):
        if not self.has_product(symbol):
            self[symbol] = None
        return self.products[symbol]

    def __iter__(self):
        return iter(self.products)

    def get_products(self):
        return self.products.values()

    def has_product(self, symbol):
        return symbol in self

    def get_product(self, symbol):
        return self[symbol]

    def set_product(self, symbol, product):
        self[symbol] = product
        return self[symbol]

    def ensure_product(self, symbol, product=None):
        if symbol not in self:
            self.set_product(symbol, product)
        return self.get_product(symbol)

    def items(self):
        return self.products.items()

    def keys(self):
        return self.products.keys()

    def values(self):
        return self.products.values()

    def place(self, order):
        if order.id in self.entries:
            raise ValueError('duplicate order id')
        product = self.ensure_product(order.symbol)
        product.place(order)
        self.entries[order.id] = product.entries[order.id]

    def cancel(self, order):
        if order.id not in self.entries:
            raise ValueError('no such order id')
        entry = self.entries[order.id]
        product = self.ensure_product(entry.symbol)
        product.cancel(order)

    def execute(self, order=None):
        if order is not None:
            product = self.ensure_product(order.symbol)
            executions = product.execute(order)
            self.entries[order.id] = product.entries[order.id]
            return executions
        else:
            executions = []
            for symbol in self.products:
                executions.extend(self.products[symbol].execute())
            return executions

    def place_order(self, *args, **kwargs):
        return self.place(Order(*args, **kwargs))

    def cancel_order(self, *args, **kwargs):
        return self.cancel(Order(*args, **kwargs))

    def execute_order(self, *args, **kwargs):
        return self.execute(Order(*args, **kwargs))

    def get_order_by_id(self, order_id):
        if order_id not in self.entries:
            return None
        entry = self.entries[order_id]
        return entry.order
